open import Data.Nat
open import Data.Bool hiding (T)
open import Data.List
open import Data.Product
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Lemmas

variable
  T : Set
  S S' S'' : List Set


----------------------------------------------------------------------
-- Source language
----------------------------------------------------------------------

data Exp : Set → Set where
  val : T → Exp T
  add : Exp ℕ → Exp ℕ → Exp ℕ
  if : Exp Bool → Exp T → Exp T → Exp T

eval : Exp T → T
eval (val x)    = x
eval (add x y)  = eval x + eval y
eval (if b x y) = if eval b then eval x else eval y


----------------------------------------------------------------------
-- Target language
----------------------------------------------------------------------

data Stack : List Set → Set where
  ε : Stack []
  _▷_ : T → Stack S → Stack (T ∷ S)

infixr 20 _▷_

data Code : List Set → List Set → Set₁ where
  PUSH : T → Code (T ∷ S) S' → Code S S'
  ADD : Code (ℕ ∷ S) S' → Code (ℕ ∷ ℕ ∷ S) S'
  IF : Code S S' → Code S S' → Code (Bool ∷ S) S'
  HALT : Code S S

exec : Code S S' → Stack S → Stack S'
exec (PUSH x c) s        = exec c (x ▷ s)
exec (IF c1 c2) (b ▷ s)  = if b then exec c1 s else exec c2 s
exec (ADD c) (m ▷ n ▷ s) = exec c ((n + m) ▷ s)
exec HALT s              = s


----------------------------------------------------------------------
-- Compiler
----------------------------------------------------------------------

comp : Exp T → Code (T ∷ S) S' → Code S S'
comp (val x) c    = PUSH x c
comp (if b x y) c = comp b (IF (comp x c) (comp y c))
comp (add x y) c  = comp x (comp y (ADD c))

compile : Exp T → Code S (T ∷ S)
compile e = comp e HALT


----------------------------------------------------------------------
-- Calculation
----------------------------------------------------------------------

correct : (c : Code (T ∷ S) S') → (e : Exp T) → (s : Stack S)
  → exec c (eval e ▷ s) ≡ exec (comp e c) s
correct c (val x) s = begin
    exec c (eval (val x) ▷ s)
  -- definition of eval
  ≡⟨ refl ⟩
    exec c (x ▷ s)
  -- define exec (PUSH x c) s = exec c (x ▷ s)
  ≡⟨ refl ⟩
    exec (PUSH x c) s
  ∎
correct c (if b x y) s = begin
    exec c (eval (if b x y) ▷ s)
  -- definition of eval
  ≡⟨ refl ⟩
    exec c ((if eval b then eval x else eval y) ▷ s)
  -- distributivity
  ≡⟨ cong (exec c) (if-distributes (λ z → z ▷ s) (eval b) (eval x) (eval y)) ⟩
    exec c (if eval b then eval x ▷ s else eval y ▷ s)
  -- distributivity
  ≡⟨ if-distributes (exec c) (eval b) (eval x ▷ s) (eval y ▷ s) ⟩
    (if eval b then exec c (eval x ▷ s) else exec c (eval y ▷ s))
  -- induction hypothesis for x
  ≡⟨ cong (λ z → if eval b then z else exec c (eval y ▷ s)) (correct c x s) ⟩
    (if eval b then exec (comp x c) s else exec c (eval y ▷ s))
  -- induction hypothesis for y
  ≡⟨ cong (λ z → if eval b then exec (comp x c) s else z) (correct c y s) ⟩
    (if eval b then exec (comp x c) s else exec (comp y c) s)
  --  define: exec (IF c1 c2) (b ▷ s) = if b then exec c1 s else exec c2 s
  ≡⟨ refl ⟩
    exec (IF (comp x c) (comp y c)) (eval b ▷ s)
  -- induction hypothesis for b
  ≡⟨ correct ((IF (comp x c) (comp y c))) b s ⟩
    exec (comp b (IF (comp x c) (comp y c))) s
  ∎
correct c (add x y) s = begin
    exec c (eval (add x y) ▷ s)
  -- definition of eval
  ≡⟨ refl ⟩
    exec c ((eval x + eval y) ▷ s)
  -- define: exec (ADD c) (m ▷ n ▷ s) = exec c ((n + m) ▷ s)
  ≡⟨ refl ⟩
    exec (ADD c) (eval y ▷ eval x ▷ s)
  -- induction hypothesis for y
  ≡⟨ correct (ADD c) y (eval x ▷ s) ⟩
    exec (comp y (ADD c)) (eval x ▷ s)
  -- induction hypothesis for x
  ≡⟨ correct (comp y (ADD c)) x s ⟩
    exec (comp x (comp y (ADD c))) s
  ∎

compile-correct : (e : Exp T) → (s : Stack S)
  → eval e ▷ s ≡ exec (compile e) s
compile-correct e s = begin
    eval e ▷ s
  ≡⟨ refl ⟩
    exec HALT (eval e ▷ s)
  ≡⟨ correct HALT e s ⟩
    exec (comp e HALT) s
  ∎
