open import Data.Bool hiding (T)
open import Data.Nat
open import Data.Nat.Properties
open import Data.List
open import Data.Maybe
open import Data.Product
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Lemmas

----------------------------------------------------------------------
-- Source language
----------------------------------------------------------------------

data Ty : Set where
  nat : Ty
  han : List Ty → List Ty → Ty

variable
  T : Ty
  a b : Bool
  S S' S'' : List Ty

data Exp : Bool → Set where
  val : ℕ → Exp false
  add : Exp a → Exp b → Exp (a ∨ b)
  throw : Exp true
  catch : Exp a → Exp b → Exp (a ∧ b)

-- Evaluation semantics where expression can throw an uncaught exception
eval? : Exp b → Maybe ℕ
eval? (val n) = just n
eval? (add x y) = maybe′ (λ n → (maybe′ (λ m → just (n + m)) nothing (eval? y))) nothing (eval? x)
eval? throw = nothing
eval? (catch x h) = maybe′ (λ n → just n) (eval? h) (eval? x)

-- Evaluation semantics where expression cannot throw an uncaught exception
eval : b ≡ false → Exp b → ℕ
eval p (val n) = n
eval p (add {false} {false} x y)  = eval refl x + eval refl y
eval p (catch {false} {a} x h) = eval refl x
eval p (catch {true} {false} x h) = maybe′ (λ n → n) (eval refl h) (eval? x)

-- Proof that, in the case where e cannot throw, eval and eval? are equivalent
eval-correct : (p : b ≡ false) → (e : Exp b) → just (eval p e) ≡ eval? e
eval-correct p (val n) = refl
eval-correct p (add {false} {false} x y) rewrite sym (eval-correct refl x) | sym (eval-correct refl y) = refl
eval-correct p (catch {false} {a} x h) rewrite sym (eval-correct refl x) = refl
eval-correct p (catch {true} {false} x h) with eval? x
eval-correct p (catch {true} {false} x h) | just n  = refl
eval-correct p (catch {true} {false} x h) | nothing = eval-correct refl h


----------------------------------------------------------------------
-- Target language
----------------------------------------------------------------------

data Code : List Ty → List Ty → Set where
  PUSH : ℕ → Code (nat ∷ S) S' → Code S S'
  ADD : Code (nat ∷ S) S'  → Code (nat ∷ nat ∷ S) S'
  THROW : Code (S'' ++ han S S' ∷ S) S'
  MARK : (h : Code S S') → (c : Code (han S S' ∷ S) S') → Code S S'
  UNMARK : Code (T ∷ S) S'  → Code (T ∷ han S S' ∷ S) S'
  HALT : Code S S

El : Ty → Set
El nat = ℕ
El (han S S') = Code S S'

data Stack : List Ty → Set where
  ε : Stack []
  _▷_ : El T → Stack S → Stack (T ∷ S)

infixr 20 _▷_

-- The definitions for exec and fail calculated in the paper do not
-- pass the totality checker, as they are not structurally
-- recursive. We fix this problem by passing around an extra arguemnt,
-- representing the size of terms, and a proof that this size is
-- greater-than or equal to the actual size of the term. This is not a
-- strict equality relationship, because it decreases at different
-- rates depending on the code constructor being evaluated.

-- The sizeX functions calculate the sizes of code/values/stacks.
sizeCode : Code S S' → ℕ
sizeCode (PUSH x c) = suc (suc (sizeCode c))
sizeCode (ADD c) = sizeCode c
sizeCode THROW = zero
sizeCode (MARK h c) = suc (suc (sizeCode c + sizeCode h))
sizeCode (UNMARK c) = suc (sizeCode c)
sizeCode HALT = zero

sizeEl : El T → ℕ
sizeEl {nat} x = 0
sizeEl {han S S'} x = sizeCode x

sizeSta : Stack S → ℕ
sizeSta ε = suc zero
sizeSta (x ▷ s) = suc (sizeEl x + sizeSta s)

-- Running UNMARK on a stack does not make it bigger
unmark-smaller : {s : Stack S}(x : El T)(h : Code S S') → sizeSta (x ▷ s) ≤ sizeSta (x ▷ h ▷ s)
unmark-smaller {T = nat} x h = s≤s (≤-step (n≤m+n _ _))
unmark-smaller {T = han _ _} x h = s≤s (+-over-≤ (≤-step (n≤m+n _ _)))

-- exec' and fail' correspond to the definition of exec and fail in
-- the paper, but with additional arguments to aid totality checking.
-- The proof argument is marked as irrelevant, so its actual value
-- doesn't matter; it just has to exist.
exec' : (c : Code S S') → (s : Stack S)
  → (size : ℕ) → .(p : size ≥ sizeCode c + sizeSta s)
  → Stack S'

fail' : (s : Stack (S'' ++ han S S' ∷ S))
  → (size : ℕ) → .(p : size ≥ sizeSta s)
  → Stack S'
fail' {[]} (h' ▷ s)                (suc size) p
  = exec' h' s                     size (≤-pred p)
fail' {_ ∷ _} (n ▷ s)              (suc size) p
  = fail' s                        size (≤-trans (n≤m+n (sizeEl n) (sizeSta s)) (≤-pred p))

exec' (PUSH n c) s                 (suc size) p
  = exec' c (n ▷ s)                size (≤-pred (≤-trans (s≤s (≤-reflexive (+-suc (sizeCode c) (sizeSta s)))) p))
exec' (ADD c) (m ▷ n ▷ s)          (suc size) p
  = exec' c ((n + m) ▷ s)          size (≤-pred (≤-trans (≤-reflexive (sym (+-suc (sizeCode c) (suc (sizeSta s))))) p))
exec' THROW s                      size p
  = fail' s                        size p
exec' (MARK h c) s                 (suc size) p
  = exec' c (h ▷ s)                size (≤-pred (≤-trans (s≤s (≤-reflexive sabc≡asbc)) p))
exec' (UNMARK c) (x ▷ h ▷ s)       (suc size) p
  = exec' c (x ▷ s)                size (≤-trans (+-over-≤ (unmark-smaller x h)) (≤-pred p))
exec' HALT s                       size p
  = s
exec' (ADD c) (m ▷ n ▷ s)          zero p
  rewrite (+-suc (sizeCode c) (suc (sizeSta s))) = s≤z-absurd p

-- exec and fail simply call exec' and fail' with the appropriate size
-- argument and proof
exec : (c : Code S S') → (s : Stack S) → Stack S'
exec c s = exec' c s (sizeCode c + sizeSta s) ≤-refl

fail : Stack (S'' ++ han S S' ∷ S) → Stack S'
fail s = fail' s (sizeSta s) ≤-refl


----------------------------------------------------------------------
-- Proofs of defining equations for exec and fail
----------------------------------------------------------------------

-- Due to the size arguemnt to exec', the definitions from the paper
-- do not automatically reduce. We have proven each of the defining
-- equations here, which are used in the calculation later.

execPUSH : ∀ n → (c : Code (nat ∷ S) S') → (s : Stack S)
  → exec (PUSH n c) s ≡ exec c (n ▷ s)
execPUSH n c s =
 begin
    exec' (PUSH n c) s (sizeCode (PUSH n c) + sizeSta s) (≤-refl)
  ≡⟨ cong₂-irr (exec' c (n ▷ s)) ((sym (+-suc (sizeCode c) (sizeSta s)))) ⟩
    exec' (PUSH n c) s (suc (sizeCode c + sizeSta (n ▷ s))) (s≤s (≤-reflexive (sym (+-suc (sizeCode c) (sizeSta s)))))
  ≡⟨ cong₂-irr (exec' c (n ▷ s)) (+-suc (sizeCode c) (sizeSta s)) ⟩
    exec' (PUSH n c) s (suc (suc (sizeCode c + sizeSta s))) (s≤s (s≤s ≤-refl))
  ≡⟨ cong₂-irr (exec' c (n ▷ s)) (sym (+-suc (sizeCode c) (sizeSta s))) ⟩
    exec' c (n ▷ s) (sizeCode c + sizeSta (n ▷ s)) ≤-refl
  ∎

execADD : ∀ n m → (c : Code (nat ∷ S) S') → (s : Stack S)
  → exec (ADD c) (m ▷ n ▷ s) ≡ exec c ((n + m) ▷ s)
execADD n m c s =
  begin
    exec (ADD c) (m ▷ n ▷ s)
  ≡⟨ refl ⟩
    exec' (ADD c) (m ▷ n ▷ s) (sizeCode (ADD c) + sizeSta (m ▷ n ▷ s)) ≤-refl
  ≡⟨ cong₂-irr (exec' (ADD c) (m ▷ n ▷ s)) {v = ≤-reflexive (+-suc (sizeCode c) (suc (sizeSta s)))}
    (+-suc (sizeCode c) (suc (sizeSta s))) ⟩
    exec' (ADD c) (m ▷ n ▷ s) (suc (sizeCode c + (suc (sizeSta s)))) (≤-reflexive (+-suc (sizeCode c) (suc (sizeSta s))))
  ≡⟨ cong₂-irr (exec' c ((n + m) ▷ s)) refl ⟩
    exec' c ((n + m) ▷ s) (sizeCode c + suc (sizeSta s)) ≤-refl
  ≡⟨ refl ⟩
    exec c ((n + m) ▷ s)
  ∎

execTHROW : (s : Stack (S'' ++ han S S' ∷ S))
  → exec THROW s ≡ fail s
execTHROW s = refl

-- Proofs that the size argument to exec and fail do not matter; the
-- proof arguemnt to these functions ensures that the values are
-- always correct
exec-size-irr : (size₁ size₂ : ℕ) → (c : Code S S') → (s : Stack S) → (p₁ : _)(p₂ : _) → exec' c s size₁ p₁ ≡ exec' c s size₂ p₂

fail-size-irr : (size₁ size₂ : ℕ) → (s : Stack (S'' ++ han S S' ∷ S)) → (p₁ : _)(p₂ : _) → fail' s size₁ p₁ ≡ fail' s size₂ p₂
fail-size-irr {[]} zero size₂ (x ▷ s) () p₂
fail-size-irr {_ ∷ _} zero size₂ (x ▷ s) () p₂
fail-size-irr {[]} (suc size₁) zero (x ▷ s) p₁ ()
fail-size-irr {_ ∷ _} (suc size₁) zero (x ▷ s) p₁ ()
fail-size-irr {[]} (suc size₁) (suc size₂) (h ▷ s) p₁ p₂ = exec-size-irr size₁ size₂ h s (≤-pred p₁) (≤-pred p₂)
fail-size-irr {_ ∷ _} (suc size₁) (suc size₂) (x ▷ s) p₁ p₂ = fail-size-irr size₁ size₂ s
  (≤-pred (≤-trans (s≤s (n≤m+n (sizeEl x) (sizeSta s))) p₁))
  (≤-pred (≤-trans (s≤s (n≤m+n (sizeEl x) (sizeSta s))) p₂))

exec-size-irr (suc size₁) (suc size₂) (PUSH x c) s p₁ p₂ = exec-size-irr size₁ size₂ c (x ▷ s)
  (x+sy (sizeCode c) (sizeSta s) size₁ p₁)
  (x+sy (sizeCode c) (sizeSta s) size₂ p₂)
exec-size-irr zero zero (ADD c) s p₁ p₂ = refl
exec-size-irr zero (suc size₂) (ADD c) (x ▷ s) p₁ p₂ = m+s≤z-absurd p₁
exec-size-irr (suc size₁) zero (ADD c) (x ▷ s) p₁ p₂ = m+s≤z-absurd p₂
exec-size-irr (suc size₁) (suc size₂) (ADD c) (m ▷ n ▷ s) p₁ p₂ = exec-size-irr size₁ size₂ c ((n + m) ▷ s)
  (n+ssm≤si p₁) (n+ssm≤si p₂)
exec-size-irr zero zero THROW s p₁ p₂ = refl
exec-size-irr zero (suc size₂) (THROW {[]}) (x ▷ s) () p₂
exec-size-irr zero (suc size₂) (THROW {_ ∷ _}) (x ▷ s) () p₂
exec-size-irr (suc size₁) zero (THROW {[]}) (x ▷ s) p₁ ()
exec-size-irr (suc size₁) zero (THROW {_ ∷ _}) (x ▷ s) p₁ ()
exec-size-irr (suc size₁) (suc size₂) THROW s p₁ p₂ = fail-size-irr (suc size₁) (suc size₂) s p₁ p₂
exec-size-irr (suc size₁) (suc size₂) (MARK h c) s p₁ p₂ = exec-size-irr size₁ size₂ c (h ▷ s)
  (≤-pred (≤-trans (s≤s (≤-reflexive sabc≡asbc)) p₁)) (≤-pred (≤-trans (s≤s (≤-reflexive sabc≡asbc)) p₂))
exec-size-irr (suc size₁) (suc size₂) (UNMARK c) (n ▷ h ▷ s) p₁ p₂ = exec-size-irr size₁ size₂ c (n ▷ s)
  (≤-pred (≤-trans (s≤s (+-over-≤ (s≤s (+-over-≤ (≤-step (n≤m+n (sizeCode h) (sizeSta s))))))) p₁))
  (≤-pred (≤-trans (s≤s (+-over-≤ (s≤s (+-over-≤ (≤-step (n≤m+n (sizeCode h) (sizeSta s))))))) p₂))
exec-size-irr size₁ size₂ HALT s p₁ p₂ = refl

execUNMARK : (c : Code (nat ∷ S) S')(h : Code S S')(n : ℕ) → (s : Stack S)
  → exec (UNMARK c) (n ▷ h ▷ s) ≡ exec c (n ▷ s)
execUNMARK c h n s = begin
    exec (UNMARK c) (n ▷ h ▷ s)
  ≡⟨ refl ⟩
    exec' (UNMARK c) (n ▷ h ▷ s) (suc (sizeCode c) + suc (suc (sizeCode h + sizeSta s))) ≤-refl
  ≡⟨ exec-size-irr (sizeCode c + suc (suc (sizeCode h + sizeSta s))) (sizeCode c + suc (sizeSta s)) c (n ▷ s)
     (+-over-≤ (s≤s (n≤m+n (suc (sizeCode h)) (sizeSta s)))) ≤-refl ⟩
    exec' c (n ▷ s) (sizeCode c + suc (sizeSta s)) ≤-refl
  ≡⟨ refl ⟩
    exec c (n ▷ s)
  ∎

execMARK : (c : Code (han S S' ∷ S) S')(h : Code S S') → (s : Stack S)
  → exec (MARK h c) s ≡ exec c (h ▷ s)
execMARK c h s = begin
    exec (MARK h c) s
  ≡⟨ refl ⟩
    exec' (MARK h c) s (suc (suc (sizeCode c + sizeCode h + sizeSta s))) ≤-refl
  ≡⟨ cong₂-irr (exec' c (h ▷ s)) (sym sabc≡asbc) ⟩
    exec' c (h ▷ s) (sizeCode c + suc (sizeCode h + sizeSta s)) ≤-refl
  ≡⟨ refl ⟩
    exec c (h ▷ s)
  ∎

failHan : (s : Stack S)(h : Code S S')
  → fail {S'' = []} (h ▷ s) ≡ exec h s
failHan s h = refl


----------------------------------------------------------------------
-- Compilers
----------------------------------------------------------------------

comp? : Exp b → Code (nat ∷ (S'' ++ han S S' ∷ S)) S' → Code (S'' ++ han S S' ∷ S) S'
comp? (val n) c = PUSH n c
comp? (add x y) c = comp? x (comp? {S'' = nat ∷ _} y (ADD c))
comp? throw c = THROW
comp? (catch x h) c = MARK (comp? h c) (comp? {S'' = []} x (UNMARK c))

comp : b ≡ false → Exp b → Code (nat ∷ S) S' → Code S S'
comp p (val x) c = PUSH x c
comp p (add {false} {false} x y) c = comp refl x (comp refl y (ADD c))
comp p (catch {false} {_} x h) c = comp refl x c
comp p (catch {true} {false} x h) c = MARK (comp refl h c) (comp? {S'' = []} x (UNMARK c))

compile : b ≡ false → Exp b → Code S (nat ∷ S)
compile p e = comp p e HALT


----------------------------------------------------------------------
-- Calculations
----------------------------------------------------------------------

correct? :
  (e : Exp b)
  → (s : Stack (S'' ++ han S S' ∷ S))
  → (c : Code ((nat ∷ S'') ++ han S S' ∷ S) S')
  → maybe′ (λ n → exec c (n ▷ s)) (fail s) (eval? e) ≡ exec (comp? e c) s
correct? (val n) s c = begin
    maybe′ (λ n' → exec c (n' ▷ s)) (fail s) (eval? (val n))
  -- definition of eval
  ≡⟨ refl ⟩
    exec c (n ▷ s)
  -- definition of exec
  ≡⟨ sym (execPUSH n c s) ⟩
    exec (PUSH n c) s
  ∎
correct? throw s c = begin
    maybe′ (λ n → exec c (n ▷ s)) (fail s) (eval? throw)
  -- definition of eval
  ≡⟨ refl ⟩
    fail s
  -- define: exec THROW s = fail s
  ≡⟨ sym (execTHROW s) ⟩
    exec THROW s
  ∎
correct? {S'' = S''} (add x y) s c = begin
    maybe′ (λ n → exec c (n ▷ s)) (fail s) (eval? (add x y))
  -- definition of eval
  ≡⟨ maybeFlatten₃ (fail s) (eval? x) (eval? y) (λ n → exec c (n ▷ s)) _+_ ⟩
    maybe′ (λ n → maybe′ (λ m → exec c ((n + m) ▷ s)) (fail s) (eval? y)) (fail s) (eval? x)
  -- definition of exec
  ≡⟨ cong (λ z → maybe′ z (fail s) (eval? x)) (funext λ n → cong (λ z → maybe′ z (fail s) (eval? y)) (funext (λ m → sym (execADD n m c s)))) ⟩
    maybe′ (λ n → maybe′ (λ m → exec (ADD c) (m ▷ n ▷ s)) (fail s) (eval? y)) (fail s) (eval? x)
  -- define: fail (n ▷ s) = fail s
  ≡⟨ cong (λ z → maybe′ (λ n → maybe′ (λ m → exec (ADD c) (m ▷ n ▷ s)) z (eval? y)) (fail s) (eval? x)) refl ⟩
    maybe′ (λ n → maybe′ (λ m → exec (ADD c) (m ▷ n ▷ s)) (fail {S'' = nat ∷ _} (n ▷ s)) (eval? y)) (fail s) (eval? x)
  -- induction hypothesis for y
  ≡⟨ cong (λ z → maybe′ z (fail s) (eval? x)) (funext (λ n → correct? {S'' = nat ∷ S''} y (n ▷ s) (ADD c))) ⟩
    maybe′ (λ n → exec (comp? y (ADD c)) (n ▷ s)) (fail s) (eval? x)
  -- induction hypothesis for x
  ≡⟨ correct? x s (comp? {S'' = nat ∷ S''} y (ADD c)) ⟩
    exec (comp? x (comp? {S'' = nat ∷ S''} y (ADD c))) s
  ∎
correct? {S'' = S''} (catch x h) s c = begin
    maybe′ (λ n → exec c (n ▷ s)) (fail s) (eval? (catch x h))
  -- definition of eval
  ≡⟨ maybeFlatten₂ (λ n → exec c (n ▷ s)) (fail s) (eval? x) (eval? h) ⟩
    maybe′ (λ n → exec c (n ▷ s)) (maybe′ (λ n → exec c (n ▷ s)) (fail s) (eval? h)) (eval? x)
  -- induction hypothesis for h
  ≡⟨ cong (λ z → maybe′ (λ n → exec c (n ▷ s)) z (eval? x)) (correct? h s c) ⟩
    maybe′ (λ n → exec c (n ▷ s)) (exec (comp? h c) s) (eval? x)
  -- define: fail {S'' = han _ _ ∷ _} (h ▷ s) = exec h s
  ≡⟨ refl ⟩
    maybe′ (λ n → exec c (n ▷ s)) (fail {S'' = []} (comp? {S'' = S''} h c ▷ s)) (eval? x)
  -- definition of exec
  ≡⟨ cong (λ z → maybe′ z (fail {[]} (comp? {S'' = S''} h c ▷ s)) (eval? x)) (funext (λ n → sym (execUNMARK c (comp? {S'' = S''} h c) n s))) ⟩
    maybe′ (λ n → exec (UNMARK c) (n ▷ comp? {S'' = S''} h c ▷ s)) (fail {S'' = []} (comp? {S'' = S''} h c ▷ s)) (eval? x)
  -- induction hypothesis for x
  ≡⟨ correct? x (comp? h c ▷ s) (UNMARK c) ⟩
    exec (comp? {S'' = []} x (UNMARK c)) (comp? {S'' = S''} h c ▷ s)
  -- definition of exec
  ≡⟨ sym (execMARK (comp? {S'' = []} x (UNMARK c)) (comp? h c) s) ⟩
    exec (MARK (comp? {S'' = S''} h c) (comp? {S'' = []} x (UNMARK c))) s
  ∎

correct : (p : b ≡ false)(e : Exp b)(s : Stack S)(c : Code (nat ∷ S) S')
  → exec c ((eval p e) ▷ s) ≡ exec (comp p e c) s
correct p (val n) s c = begin
    exec c (eval p (val n) ▷ s)
  -- definition of eval
  ≡⟨ refl ⟩
    exec c (n ▷ s)
  -- define: exec (PUSH n c) s = exec c (n ▷ s)
  ≡⟨ sym (execPUSH n c s) ⟩
    exec (PUSH n c) s
  ∎
correct p (add {false} {false} x y) s c = begin
    exec c (eval p (add x y) ▷ s)
  -- definition of eval
  ≡⟨ refl ⟩
    exec c ((eval refl x + eval refl y) ▷ s)
  -- define: exec (ADD c) (m ▷ n ▷ s) = exec c (n + m ▷ s)
  ≡⟨ sym (execADD (eval refl x) (eval refl y) c s) ⟩
    exec (ADD c) (eval refl y ▷ eval refl x ▷ s)
  -- inductive hypothesis for y
  ≡⟨ correct refl y (eval refl x ▷ s) (ADD c) ⟩
    exec (comp refl y (ADD c)) (eval refl x ▷ s)
  -- inductive hypothesis for x
  ≡⟨ correct refl x s (comp refl y (ADD c)) ⟩
    exec (comp refl x (comp refl y (ADD c))) s
  ∎
correct p (catch {false} {_} x h) s c = begin
    exec c (eval p (catch x h) ▷ s)
  -- aplying eval
  ≡⟨ refl ⟩
    exec c (eval refl x ▷ s)
  -- inductive hypothesis for x
  ≡⟨ correct refl x s c ⟩
    exec (comp refl x c) s
  ∎
correct p (catch {true} {false} x h) s c = begin
    exec c (eval p (catch x h) ▷ s)
  -- definition of eval
  ≡⟨ refl ⟩
    exec c (maybe′ (λ n → n) (eval refl h) (eval? x) ▷ s)
  -- distributivity of case
  ≡⟨ maybeDistributes (λ n → exec c (n ▷ s)) (λ n → n) (eval refl h) (eval? x) ⟩
    maybe′ (λ n → exec c (n ▷ s)) (exec c (eval refl h ▷ s)) (eval? x)
  -- inductive hypothesis for h
  ≡⟨ cong (λ z → maybe′ (λ n → exec c (n ▷ s)) z (eval? x)) (correct refl h s c) ⟩
    maybe′ (λ n → exec c (n ▷ s)) (exec (comp refl h c) s) (eval? x)
  -- define: fail {S'' = han _ _ ∷ _} (h ▷ s) = exec h s
  ≡⟨ cong (λ z → maybe′ (λ n → exec c (n ▷ s)) z (eval? x)) (sym (failHan s (comp refl h c))) ⟩
    maybe′ (λ n → exec c (n ▷ s)) (fail {[]} (comp refl h c ▷ s)) (eval? x)
  -- define: exec (UNMARK c) (x ▷ h ▷ s) = exec c (x ▷ s)
  ≡⟨ cong (λ z → maybe′ z (fail {[]} (comp refl h c ▷ s)) (eval? x)) (funext λ n → sym (execUNMARK c (comp refl h c) n s)) ⟩
    maybe′ (λ n → exec (UNMARK c) (n ▷ comp refl h c ▷ s)) (fail {[]} (comp refl h c ▷ s)) (eval? x)
  -- specification (4) for x
  ≡⟨ correct? {S'' = []} x (comp refl h c ▷ s) (UNMARK c) ⟩
    exec (comp? x (UNMARK c)) (comp refl h c ▷ s)
  -- define: exec (MARK h c) s = exec c (h ▷ s)
  ≡⟨ sym (execMARK (comp? {S'' = []} x (UNMARK c)) (comp refl h c) s) ⟩
    exec (MARK (comp refl h c) (comp? {S'' = []} x (UNMARK c))) s
  ∎

compile-correct : (p : b ≡ false) → (e : Exp b) → (s : Stack S)
  → eval p e ▷ s ≡ exec (compile p e) s
compile-correct p e s = begin
    eval p e ▷ s
  ≡⟨ refl ⟩
    exec HALT (eval p e ▷ s)
  ≡⟨ correct p e s HALT ⟩
    exec (comp p e HALT) s
  ∎
