open import Data.Nat
open import Data.Bool hiding (T)
open import Data.List
open import Data.Product
open import Relation.Binary.PropositionalEquality
open ≡-Reasoning

open import Lemmas

variable
  T : Set
  S S' S'' : List Set


----------------------------------------------------------------------
-- Source language
----------------------------------------------------------------------

data Exp : Set → Set where
  val : T → Exp T
  add : Exp ℕ → Exp ℕ → Exp ℕ
  if : Exp Bool → Exp T → Exp T → Exp T

eval : Exp T → T
eval (val x)    = x
eval (add x y)  = eval x + eval y
eval (if b x y) = if eval b then eval x else eval y


----------------------------------------------------------------------
-- Target language
----------------------------------------------------------------------

data Stack : List Set → Set where
  ε : Stack []
  _▷_ : T → Stack S → Stack (T ∷ S)

infixr 20 _▷_

data Code : List Set → List Set → Set₁ where
  PUSH  : T → Code S (T ∷ S)
  ADD   : Code (ℕ ∷ ℕ ∷ S) (ℕ ∷ S)
  _+++_ : Code S S' → Code S' S'' → Code S S''
  IF    : Code S S' → Code S S' → Code (Bool ∷ S) S'

exec : Code S S' → Stack S → Stack S'
exec (PUSH x) s         = x ▷ s
exec (IF c1 c2) (b ▷ s) = if b then exec c1 s else exec c2 s
exec (c1 +++ c2) s      = exec c2 (exec c1 s)
exec ADD (m ▷ n ▷ s)    = (n + m) ▷ s


----------------------------------------------------------------------
-- Compiler
----------------------------------------------------------------------

compile : Exp T → Code S (T ∷ S)
compile (val x)    = PUSH x
compile (if b x y) = compile b +++ IF (compile x) (compile y)
compile (add x y)  = (compile x +++ compile y) +++ ADD


----------------------------------------------------------------------
-- Calculation
----------------------------------------------------------------------

correct : (e : Exp T) → (s : Stack S) → eval e ▷ s ≡ exec (compile e) s
correct (val x) s = begin
    eval (val x) ▷ s
  -- definition of eval
  ≡⟨ refl ⟩
    x ▷ s
  -- define exec (PUSH x) s = x ▷ s
  ≡⟨ refl ⟩
    exec (PUSH x) s
  ∎
correct (if b x y) s = begin
    eval (if b x y) ▷ s
  -- definition of eval
  ≡⟨ refl ⟩
    (if eval b then eval x else eval y) ▷ s
  -- distributivity
  ≡⟨ if-distributes (λ z → z ▷ s) (eval b) (eval x) (eval y) ⟩
    (if eval b then eval x ▷ s else eval y ▷ s)
  -- induction hypothesis for x
  ≡⟨ cong (λ z → if eval b then z else eval y ▷ s) (correct x s) ⟩
    (if eval b then exec (compile x) s else eval y ▷ s)
  -- induction hypothesis for y
  ≡⟨ cong (λ z → if eval b then exec (compile x) s else z) (correct y s) ⟩
    (if eval b then exec (compile x) s else exec (compile y) s)
  -- define: exec (IF c1 c2) (b ▷ s) = if b then exec c1 s else exec c2 s
  ≡⟨ refl ⟩
    exec (IF (compile x) (compile y)) (eval b ▷ s)
  -- define: exec (c1 +++ c2) = exec c2 (exec c1 s)
  ≡⟨ cong (exec (IF (compile x) (compile y))) (correct b s) ⟩
    exec (IF (compile x) (compile y)) (exec (compile b) s)
  -- define: exec (c1 +++ c2) = exec c2 (exec c1 s)
  ≡⟨ refl ⟩
    exec (compile b +++ IF (compile x) (compile y)) s
  ∎
correct (add x y) s = begin
    eval (add x y) ▷ s
  -- definition of eval
  ≡⟨ refl ⟩
    (eval x + eval y) ▷ s
  -- define exec ADD (m ▷ n ▷ s) = (n + m) ▷ s
  ≡⟨ refl ⟩
    exec ADD (eval y ▷ eval x ▷ s)
  -- induction hypothesis for x
  ≡⟨ cong (λ z → exec ADD (eval y ▷ z)) (correct x s) ⟩
    exec ADD (eval y ▷ exec (compile x) s)
  -- induction hypothesis for y
  ≡⟨ cong (exec ADD) (correct y (exec (compile x) s)) ⟩
    exec ADD (exec (compile y) (exec (compile x) s))
  -- definition of +++
  ≡⟨ refl ⟩
    exec ADD (exec (compile x +++ compile y) s)
  -- definition of +++
  ≡⟨ refl ⟩
    exec ((compile x +++ compile y) +++ ADD) s
  ∎
