open import Data.Empty
open import Data.Fin hiding (lift)
open import Data.List hiding (lookup)
open import Data.Nat
open import Data.Product
open import Data.Vec
open import Relation.Binary.PropositionalEquality


----------------------------------------------------------------------
-- Types and variables
----------------------------------------------------------------------

-- whether a value is used in the setting of expressions or compiled code
data Setting : Set where
  E : Setting
  C : Setting

variable
  n : ℕ
  s : Setting
  T : Set

mutual
  StackTy : Set₁
  StackTy = List ExpTy

  Env : (n : ℕ) → Set₁
  Env n = Vec ExpTy n

  data ExpTy : Set₁ where
    fn : (a : ExpTy) → (b : ExpTy) → ExpTy
    ty : (T : Set) → ExpTy
    cont : Env n → StackTy → StackTy → ExpTy

data Lift {l} : Set → Set l where
  lift : T → Lift T

variable
  A B : ExpTy
  S S' S'' S0 S1 S2 : StackTy
  env env' env'' : Env n


----------------------------------------------------------------------
-- Source language
----------------------------------------------------------------------

data STLC : Env n → ExpTy → Set₁ where
  val : T → STLC env (ty T)
  var : (i : Fin n) → STLC env (lookup env i)
  lam : (STLC (A Vec.∷ env) B) → STLC env (fn A B)
  app : STLC env (fn A B) → STLC env A → STLC env B


----------------------------------------------------------------------
-- Target language
----------------------------------------------------------------------

data Code : Env n → StackTy → StackTy → Set₁ where
  HALT : Code env S S
  LOOKUP : (i : Fin n) → Code env (lookup env i ∷ S) S' → Code env S S'
  PUSH : (x : T) → Code env (ty T ∷ S) S' → Code env S S'
  ABS : ({S S' : StackTy}{n : ℕ}{env' : Env n}
    → Code (A Vec.∷ env) (cont env' (B ∷ S) S' ∷ S) S')
    → Code env (fn A B ∷ S) S'
    → Code env S S'
  APPLY : Code env (B ∷ S) S' → Code env (A ∷ fn A B ∷ S) S'
  RET : {env' : Env n} → Code env (A ∷ cont env' (A ∷ S) S' ∷ S) S'


----------------------------------------------------------------------
-- Compiler
----------------------------------------------------------------------

comp : STLC env A → Code env (A ∷ S) S' → Code env S S'
comp (val x) c = PUSH x c
comp (var i) c = LOOKUP i c
comp {S = S} {S' = S'} (lam e) c = ABS (comp e RET) c
comp (app x y) c = comp x (comp y (APPLY c))


----------------------------------------------------------------------
-- Semantics
----------------------------------------------------------------------

mutual
  Val : Setting → ExpTy → Set₁
  Val C (fn a b) = {S S' : StackTy}{n : ℕ}{env' : Env n} →
    Σ ℕ (λ n → Σ (Env n) (λ env → Store C env × Code (a Vec.∷ env) (cont env' (b ∷ S) S' ∷ S) S'))
  Val E (fn a b) =
    Σ ℕ (λ n → Σ (Env n) (λ env → Store E env × STLC (a Vec.∷ env) b))
  Val C (cont env S S') = Clo env S S'
  Val E (cont env S S') = Lift ⊥
  Val s (ty T) = Lift T

  data Store : Setting → Env n → Set₁ where
    εₛ : Store s Vec.[]
    _▷ₛ_ : {A : ExpTy} → (a : Val s A) → Store s env → Store s (A Vec.∷ env)

  data Clo : Env n → StackTy → StackTy → Set₁ where
    clo : Store C env → Code env S S' → Clo env S S'

variable
  sto : Store C env

get : (i : Fin n) → Store s env → Val s (lookup env i)
get zero (x ▷ₛ s) = x
get (suc i) (x ▷ₛ s) = get i s

-- evaluation semantics for the simply-typed lambda calculus
data _⇓[_]_ : STLC env A → Store E env → Val E A → Set₁ where
  evalVal : {x : T}{sto : Store E env}
    → val x ⇓[ sto ] lift x
  evalVar : {i : Fin n}{sto : Store E env}
    → var i ⇓[ sto ] get i sto
  evalLam : {n : ℕ}{env : Env n}{x : STLC (A Data.Vec.∷ env) B}{sto : _}
    → lam x ⇓[ sto ] (n , env , sto , x)
  evalApp : {sto : Store E env}{sto' : Store E env'}
    {x : STLC env (fn A B)}{y : STLC env A}{y' : _}{e : STLC (A Vec.∷ env') B}{xy : _}
    → x ⇓[ sto ] (n , env' , sto' , e)
    → y ⇓[ sto ] y'
    → e ⇓[ y' ▷ₛ sto' ] xy
    → app x y ⇓[ sto ] xy

data Stack : StackTy → Set₁ where
  ε : Stack []
  _▷_ : Val C A → Stack S → Stack (A ∷ S)

variable
  sta : Stack S

mutual
  valExpToCode : Val E A → Val C A
  valExpToCode {fn A A₁} (n , env , sto , e) = n , env , (storeExpToCode sto) , (comp e RET)
  valExpToCode {ty T} x = x
  valExpToCode {cont x₁ x₂ x₃} (lift ())

  storeExpToCode : Store E env → Store C env
  storeExpToCode εₛ = εₛ
  storeExpToCode (a ▷ₛ x) = (valExpToCode a) ▷ₛ (storeExpToCode x)

-- small-step execution semantics of code
data _,_,_==>_,_,_ : Code env S S'' → Store C env → Stack S
  → Code env' S' S'' → Store C env' → Stack S' → Set₁ where
  execHALT : HALT , sto , sta ==> HALT , sto , sta
  execLOOKUP : {i : Fin n}{c : Code env (lookup env i ∷ S) S'}
    → LOOKUP i c , sto , sta ==> c , sto , (get i sto ▷ sta)
  execPUSH : {x : T}{c : Code env (ty T ∷ S) S'}
    → PUSH x c , sto , sta ==> c , sto , (lift x ▷ sta)
  execABS : {x : {S S' : StackTy}{n : ℕ}{env' : Env n} → Code (A Vec.∷ env) (cont env' (B ∷ S) S' ∷ S) S'}{c : Code env (fn A B ∷ S) S'}
    → ABS x c , sto , sta ==> c , sto , ((_ , (_ , (sto , x))) ▷ sta)
  execAPPLY : {c : Code env (B ∷ S) S'}{v : Val C A}{f : Val C (fn A B)}
    → {sto : Store C env}
    → {sta : Stack S}
    → {sto' : Store C env'}
    → {c' : Code (A Vec.∷ env') (cont env (B ∷ S) S' ∷ S) S'}
    → f {S} {S'} ≡ (n , env' , sto' , c')
    → APPLY c , sto , (v ▷ (f ▷ sta)) ==> c' , (v ▷ₛ sto') , (clo sto c ▷ sta)
  execRET : {S S' : StackTy}{sta : Stack S}{x : Val C A}{sto' : Store C env'}
    → {c' : Code env' (A ∷ S) S'}
    → RET , sto , (x ▷ (clo sto' c' ▷ sta)) ==> c' , sto' , (x ▷ sta)

-- big-step execution semantics of code
data _,_,_=>>_,_,_ : Code env S S'' → Store C env → Stack S
  → Code env' S' S'' → Store C env' → Stack S' → Set₁ where
  execRefl : {c : Code env S S'}{sto : Store C env}{sta : Stack S} → c , sto , sta =>> c , sto , sta
  execTrans : {c : Code env S0 S}{sto : Store C env}{sta : Stack S0}
    {c' : Code env' S1 S}{sto' : Store C env'}{sta' : Stack S1}
    {c'' : Code env'' S2 S}{sto'' : Store C env''}{sta'' : Stack S2}
    → c , sto , sta ==> c' , sto' , sta'
    → c' , sto' , sta' =>> c'' , sto'' , sta''
    → c , sto , sta =>> c'' , sto'' , sta''

execLift : {c : Code env S S''}{sto : Store C env}{sta : Stack S}
    {c' : Code env' S' S''}{sto' : Store C env'}{sta' : Stack S'}
    → c , sto , sta ==> c' , sto' , sta'
    → c , sto , sta =>> c' , sto' , sta'
execLift p = execTrans p execRefl

execTrans' : {c : Code env S0 S}{sto : Store C env}{sta : Stack S0}
    {c' : Code env' S1 S}{sto' : Store C env'}{sta' : Stack S1}
    {c'' : Code env'' S2 S}{sto'' : Store C env''}{sta'' : Stack S2}
    → c' , sto' , sta' =>> c'' , sto'' , sta''
    → c , sto , sta =>> c' , sto' , sta'
    → c , sto , sta =>> c'' , sto'' , sta''
execTrans' y execRefl = y
execTrans' z (execTrans x y) = execTrans x (execTrans' z y)


----------------------------------------------------------------------
-- Calculation
----------------------------------------------------------------------

getConvStore : (i : Fin n)(sto : Store E env)(sta : Stack S)
  → (get i (storeExpToCode sto) ▷ sta)
  ≡ (valExpToCode (get i sto) ▷ sta)
getConvStore zero (a ▷ₛ sto) sta = refl
getConvStore (suc i) (a ▷ₛ sto) sta = getConvStore i sto sta

calculation :
  (e : STLC env A)
  → {e' : _}
  → (sto : Store E env)
  → (sta : Stack S)
  → (c : Code env (A ∷ S) S')
  → e ⇓[ sto ] e'
  → comp e c , storeExpToCode sto , sta =>> c , storeExpToCode sto , (valExpToCode e' ▷ sta)
calculation (val i) sto sta c evalVal = execTrans'
  -- exec c sto (eval (val x) ▷ sta)
  --   apply eval
  execRefl
  -- exec c sto (lift x ▷ sta)
  --   define exec (PUSH x c) sto sta = exec c sto (lift x ▷ sta)
  (execLift execPUSH)
  -- exec (PUSH i c) sto sta
calculation (var i) sto sta c evalVar
  rewrite sym (getConvStore i sto sta) = execTrans'
  -- exec c sto (eval sto (var i) ▷ sta)
  --   apply eval
  execRefl
  -- exec c sto (get i sto ▷ sta)
  --   define exec (LOOKUP i c) sto sta = exec c (get i sto ▷ sta)
  (execLift execLOOKUP)
  -- exec (LOOKUP i c) sto sta
calculation (lam e) sto sta c evalLam = execTrans'
  -- exec c sto (eval sto (lam e) ▷ sta)
  --   apply eval
  execRefl
  -- exec c sto (valExpToCode (n , env , sto , e) ▷ sta)
  --   apply valExpToCode
  -- exec c sto (n , env , storeExpToCode sto , comp e RET ▷ sta)
  --   define exec {n}{env} (ABS x c) sto sta = exec c sto ((n , env , sto , x) ▷ sta)
  (execLift execABS)
  -- exec (ABS (comp e RET) c) sto sta
calculation (app x y) sto sta c
  (evalApp {sto' = sto'}{y' = y'}{e = e}{xy = xy} evalx evaly evale)
  with calculation x sto sta (comp y (APPLY c)) evalx
  | calculation y sto (_ ▷ sta) (APPLY c) evaly
  | calculation e (y' ▷ₛ sto') (clo (storeExpToCode sto) c ▷ sta) RET evale
calculation (app x y) sto sta c (evalApp evalx evaly evale)
  | ihX | ihY | ihE =
  -- exec c sto (eval sto (app x y) ▷ sta)
  --   apply eval
  -- exec c sto (xy ▷ sta)
  --   define exec RET sto (x ▷ (clo sto' c' ▷ sta)) = exec c' sto' (x ▷ sta)
  execTrans' (execLift execRET)
  -- exec RET (valExpToCode y' ▷ₛ sto') (xy ▷ clo sto c ▷ sta)
  --   induction hypothesis for e
  (execTrans' ihE
  -- exec (comp e RET) (valExpToCode y' ▷ₛ sto') (clo sto c ▷ sta)
  --   define exec (APPLY c) sto (v ▷ ((n , env' , sto' , c') ▷ sta)) = exec c' (v ▷ₛ sto') (clo sto c ▷ sta)
  (execTrans' (execLift (execAPPLY refl))
  -- exec (APPLY c) sto (valExpToCode y' ▷ₛ ((n , env' , sto' , comp e RET) ▷ sta))
  --   induction hypothesis for y
  (execTrans' ihY
  -- exec (comp y (APPLY c)) (storeExpToCode sto) (valExpToCode y' ▷ₛ (n , env' , sto' , comp e RET) ▷ sta)
  --   induction hypothesis for x
  ihX
  -- exec (comp x (comp y (APPLY c))) (storeExpToCode sto) sta
  )))
